#include "dma_simple.h"

#include        <time.h>
#include        <sys/time.h>
void print_diff_time(struct timeval start_time, struct timeval end_time)
{
    struct timeval diff_time;
    if (end_time.tv_usec < start_time.tv_usec) {
        diff_time.tv_sec  = end_time.tv_sec  - start_time.tv_sec  - 1;
        diff_time.tv_usec = end_time.tv_usec - start_time.tv_usec + 1000*1000;
    } else {
        diff_time.tv_sec  = end_time.tv_sec  - start_time.tv_sec ;
        diff_time.tv_usec = end_time.tv_usec - start_time.tv_usec;
    }
    printf("time = %ld.%06ld sec\n", diff_time.tv_sec, diff_time.tv_usec);
}

void main() {
  // HLS_accel
  int uio1_fd = open("/dev/uio1", O_RDWR);
  void *hls_regs = mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, uio1_fd, 0);
  // AXI DMA
  int uio2_fd = open("/dev/uio2", O_RDWR);
  void *dma_regs = mmap(NULL, 0x1000, PROT_READ|PROT_WRITE, MAP_SHARED, uio2_fd, 0);

  // udmabuf
  struct udmabuf intake_buf;
  struct udmabuf outlet_buf;
  if (udmabuf_open(&intake_buf, "udmabuf0") == -1) exit(1);
  if (udmabuf_open(&outlet_buf, "udmabuf1") == -1) exit(1);

  dma_reset(dma_regs);   // Reset AXI DMA
  regs_write32(hls_regs, 0x81);  // Auto Start HLS_accel

  float a[32*32];
  float b[32*32];
  float c[32*32];
  float cref[32*32];
  int i, j, t;
  int m, n, k;
  m = n = k = 32;

  for(i = 0; i < m*m; i++) {
    a[i] = (float) i/2.0;
    b[i] = (float) i/3.0;
    c[i] = (float) 0;
  }

  for(i = 0; i < m; i++) {
    for(j = 0; j < n; j++) {
      ((float *)(intake_buf.buf))[i*n+j] = a[i*n+j];
    }
  }
  for(i = 0; i < m; i++) {
    for(j = 0; j < n; j++) {
      ((float *)(intake_buf.buf))[i*n+j+m*n] = b[i*n+j];
    }
  }

  struct timeval start_time, end_time;
  gettimeofday(&start_time, NULL);
  // Setting up DMA
  dma_setup(dma_regs, intake_buf.phys_addr, outlet_buf.phys_addr);

  // Starting DMA
  dma_outlet_start(dma_regs, m*m*4);
  dma_intake_start(dma_regs, m*m*2*4);

  // Wait IRQ
  dma_wait_irq(dma_regs);
  dma_clear_status(dma_regs);
  gettimeofday(&end_time, NULL);
  print_diff_time(start_time, end_time);

  for(i = 0; i < m; i++) {
    for(j = 0; j < n; j++) {
      c[i*n+j] = ((float*)(outlet_buf.buf))[i*n+j];
    }
  }

  for(i = 0; i < m; i++) {
    for(j = 0; j < n; j++) {
      float sum = 0;
      for(t = 0; t < k; t++) {
	sum += a[i*k+t] * b[t*n+j];
      }
      cref[i*n+j] = sum;
    }
  }

  // Compare
  int cnt_ok=0;
  int cnt_ng=0;
  for(i = 0; i < m*m; i++) {
    if(c[i] == cref[i]) {
      cnt_ok++;
    } else {
      printf("NG: hw %f, sw %f\n", c[i], cref[i]);
      cnt_ng++;
    }
  }
  printf("OK: %d, NG: %d, Total: %d\n", cnt_ok, cnt_ng, m*m);

  udmabuf_close(&outlet_buf);
  udmabuf_close(&intake_buf);
  close(uio2_fd);
  close(uio1_fd);
}

